/*-----------------------------------------------------------------------------
 *  Copyright (c) 1993      TRIUMF Data Acquistion Group
 *  Please leave this header in any reproduction of that distribution
 * 
 *  TRIUMF Data Acquisition Group, 4004 Wesbrook Mall, Vancouver, B.C. V6T 2A3
 *  Email: online@triumf.ca           Tel: (604) 222-1047  Fax: (604) 222-1074
 *         amaudruz@triumf.ca
 * -----------------------------------------------------------------------------
 *  
 *  Description	: CAMAC functions for the CES8210.
 *		  Inline function for interrupts. Vector 85 and 170.
 *		  Non-standard CAMAC test functions for cam8210.h
 *		  Private section for NMR readout
 *
 *                For ESONE CAMAC standard call check camac_ces8210.c
 *
 *  Requires 	: cam8210.h
 *     
 *  Author:  Pierre-Andre Amaudruz Data Acquisition Group
 * 
 *  Revision 1.0  1996/Oct/2 Pierre	 Initial revision
 *
 *---------------------------------------------------------------------------*/

#include "midas.h"
#include "msystem.h"
#include "cam8210.h"

extern excStub;

/************ INLINE function for Interrupt handling **********/
inline void isr_routine (DWORD vector)
{
  logMsg("isr_routine %i\n",vector,0,0,0,0,0);
}

inline void int_ces_clr(void)
{
  DWORD cbd_ifr;
  WORD ifr;
  
  cbd_ifr = IFR;			/* address: interrupt flag register */
  ifr = 0;
  *(WORD *)cbd_ifr = ifr;
}

inline void int_ces_enable (DWORD intnum)
{
  DWORD cbd_csr_r, cbd_ifr;
  WORD csr;
  cbd_csr_r= CSR;			/* address: camac control/status reg */
  csr = *(WORD *)cbd_csr_r;	/* read csr */
  if (intnum == 2)
    csr = csr & ~MIT2;
  else if (intnum == 4)
    csr = csr & ~MIT4;
  *(WORD *)cbd_csr_r = csr;
  sysIntEnable(intnum);	        /* interrupt level x */
}

inline void int_ces_disable (DWORD intnum)
{
  DWORD cbd_csr_r, cbd_ifr;
  WORD csr;
  cbd_csr_r= CSR;		/* address: camac control/status reg */
  csr = *(WORD *)cbd_csr_r;	/* read csr */
  if (intnum == 2)
    csr = csr | MIT2;
  else if (intnum == 4)
    csr = csr | MIT4;
  *(WORD *)cbd_csr_r = csr;
  sysIntDisable(intnum);	        /* interrupt level x */
}

inline void int_ces_book (DWORD intnum, DWORD isr_routine )
{
  INT status;

  int_ces_clr();		/* clear any pending front panel interrupts */
  int_ces_disable(intnum);	/* mask 8210 external interrupt 2 */
  sysIntDisable(intnum);	/* disable bus interrupt level 2 */
  
  if (intnum == 2)
    status=intConnect((VOIDFUNCPTR *) (85<<2),    (VOIDFUNCPTR) isr_routine,85);
  if (intnum == 4)
    status=intConnect((VOIDFUNCPTR *) (170<<2),   (VOIDFUNCPTR) isr_routine,170);
}

inline void int_ces_unbook (DWORD intnum)
{
  int status;
  clrInt();		        /* clear any pending front panel interrupts */
  int_ces_disable(intnum);		/* mask 8210 external interrupt x */
  sysIntDisable(intnum);	/* disable bus interrupt level x */
  
  if (intnum == 2)
    status=intConnect((VOIDFUNCPTR *) (85<<2),    (VOIDFUNCPTR) excStub,85);
  if (intnum == 4)
    status=intConnect((VOIDFUNCPTR *) (170<<2),   (VOIDFUNCPTR) excStub,170);
}


/************** Non Standard CAMAC call functions using the cam8210.h 
		for test purpose. Display result on VME console screen **********/
void cio(void)
{
  printf("---> CAM8210 Macros (cam8210.h) <---\n");
  printf("camop ()                 <--- Camac Status CSR, CAR, BTB\n");
  printf("btb  ()                  <--- Timeout & Online Branch  Register\n");
  printf("car  ()                  <--- Crate Address Register\n");
  printf("csr  ()                  <--- Camac Status Register\n");
  printf("cchk (b,c)               <--- Camac Crate check\n");       
  printf("ci16 (b,c,n,a,f)         <--- Camac Input 16bits\n");
  printf("ci24 (b,c,n,a,f)         <--- Camac Input 24bits\n");
  printf("co16 (b,c,n,a,f,d)       <--- Camac Output 16bits\n");
  printf("co24 (b,c,n,a,f,d)       <--- Camac Output 24bits\n");
  printf("ccom (b,c,n,a,f)         <--- Camac Command\n");
  printf("ccmd (b,c,n,a,f)         <--- Camac Command\n");
  printf("ccas (b,c,n,a,f,r)       <--- Camac Command scan on A\n");
  printf("ccns (b,c,n,a,f,r)       <--- Camac Command scan on N\n");
  printf("int_ces_clr    ()\n");
  printf("int_ces_book   ( intnum [2 or 4], function_name )\n");
  printf("int_ces_enable ( intnum [2 or 4])\n");
  printf("int_ces_disable( intnum [2 or 4])\n");
  printf("int_ces_unbook ( intnum [2 or 4])\n");
}

void csr(void)
{
  WORD online;

  CAM_CSRCHK(&online);
  printf("CSR Camac  status 0x%4.4x\n",online);
}

void car(void)
{
  WORD online;

  CAM_CARCHK(&online);
  printf("CAR Online status 0x%4.4x\n",online);
}

void btb(void)
{
  WORD online;

  CAM_BTBCHK(&online);
  printf("BTB Online status 0x%4.4x\n",online);
}

void cchk (DWORD b, DWORD c)
{
  DWORD ext;

  CAM_BC(&ext,b,c);
  printf ("Branch %i  Crate %i  ext %8.8x\n",b,c,ext);
}

void camop(void)
{
  csr();
  car();
  btb();
}

void ci16(DWORD b,DWORD c,DWORD n,DWORD a,DWORD f)
{
  DWORD ext, q, x;
  WORD d;

  CAM_BC(&ext,b,c);
  CAM16_IQX(&ext,n,a,f,&d,&q,&x);
  printf("Q%1i / X%1i - B%1iC%1iN%2.2iA%2.2iF%2.2i -> ($%8.8x) => 0x%4.4x - %i\n",q,x,b,c,n,a,f,ext,d,d); 
}

void ci24(DWORD b,DWORD c,DWORD n,DWORD a,DWORD f)
{
  DWORD ext;
  DWORD d, q, x;
  CAM_BC(&ext,b,c);
  CAM24_IQX(&ext,n,a,f,&d,&q,&x);
  printf("Q%1i / X%1i - B%1iC%1iN%2.2iA%2.2iF%2.2i -> ($%8.8x) => 0x%8.8x - %i\n",q,x,b,c,n,a,f,ext,d,d); 
}

void co16(DWORD b,DWORD c,DWORD n,DWORD a,DWORD f, WORD d)
{
  DWORD ext, q, x;

  CAM_BC(&ext,b,c);
  CAM16_OQX(&ext,n,a,f,d,&q,&x);
  printf("Q%1i / X%1i - B%1iC%1iN%2.2iA%2.2iF%2.2i -> ($%8.8x) <= 0x%4.4x - %i\n",q,x,b,c,n,a,f,ext,d,d); 
}

void co24(DWORD b,DWORD c,DWORD n,DWORD a,DWORD f, DWORD d)
{
  DWORD ext, q, x;

  CAM_BC(&ext,b,c);
  CAM24_OQX(&ext,n,a,f,d,&q,&x);
  printf("Q%1i / X%1i - B%1iC%1iN%2.2iA%2.2iF%2.2i -> ($%8.8x) <= 0x%8.8x - %i\n",q,x,b,c,n,a,f,ext,d,d); 
}

void ccom(DWORD b,DWORD c,DWORD n,DWORD a,DWORD f)
{
  DWORD q, x;

  CAM_COM(b,c,n,a,f);
  CAM_QXCHK(&q,&x);           
  printf("Q%1i / X%1i - B%1iC%1iN%2.2iA%2.2iF%2.2i -> () <= action\n",q,x,b,c,n,a,f); 
}

void ccmd(DWORD b,DWORD c,DWORD n,DWORD a,DWORD f)
{
  DWORD ext, q, x;

  CAM_BC(&ext,b,c);
  CAM_CMD(&ext,n,a,f);
  CAM_QXCHK(&q,&x);           
  printf("Q%1i / X%1i - B%1iC%1iN%2.2iA%2.2iF%2.2i -> ($%8.8x) <= action\n",q,x,b,c,n,a,f,ext); 
}

void ccas(DWORD b,DWORD c,DWORD n,DWORD a,DWORD f,DWORD r)
{
  DWORD ext, q, x;

  CAM_BC(&ext,b,c);
  CAM_CMDAS(&ext,n,a,f,r);
  CAM_QXCHK(&q,&x);
  printf("Q%1i / X%1i - B%1iC%1iN%2.2iA%2.2iF%2.2i -> ($%8.8x) <= action\n",q,x,b,c,n,a,f,ext); 
}

void ccns(DWORD b,DWORD c,DWORD n,DWORD a,DWORD f,DWORD r)
{
  DWORD ext, q, x;

  CAM_BC(&ext,b,c);
  CAM_CMDNS(&ext,n,a,f,r);
  CAM_QXCHK(&q,&x);
  printf("Q%1i / X%1i - B%1iC%1iN%2.2iA%2.2iF%2.2i -> (%8.8x) <= action\n",q,x,b,c,n,a,f,ext); 
}

/*-- Private B12 NMR readout function where data are in ASCII format... */
float read_b12 (DWORD subadd)
{
  DWORD d, lval2, dat1,dat2,dat3,dat4,dat5,dat6;
  char  str[8];
  float value;
  DWORD ext_crate2;
  
  CAM_BC(&ext_crate2,0,2);
  CAM24_I(&ext_crate2,9,subadd,0,&d);  
  lval2 = ~d;
  if (((lval2 >>3) & 1) == 0)
    { 
      dat1 = lval2 & 0x7;
      str[0] = '0' + dat1;
      dat2 = ((lval2 >> 4 ) & 0x0F);
      str[1] = '0' + dat2;
      dat3 = ((lval2 >> 8 ) & 0x0F);
      str[2] = '0' + dat3;
      dat4 = ((lval2 >> 12) & 0x0F);
      str[3] = '0' + dat4;
      dat5 = ((lval2 >> 16) & 0x0F);
      str[4] = '0' + dat5;
      dat6 = ((lval2 >> 20) & 0x0F);
      str[5] = '0' + dat6;
      str[6] = '\0';
      value = atof (str) / 10.;  /* [Gauss] */
   }
  else
   {
      value = 0.0;
    }
  printf("value %f\n",value);
  return (value);
}	

