/*********************************************************************

  Name:         lrs2373.c
  Created by:   Stefan Ritt

  Cotents:      Routines for LeCroy 2365 Octal Logic Matrix
                

  Revision history
  ------------------------------------------------------------------
  date        by    modification
  ---------   ---   ------------------------------------------------
  15-MAY-97   SR    created

*********************************************************************/

#include <stdio.h>
#include <dos.h>
#include <string.h>

#include "midas.h"

#define CAMAC_BASE_DEFINED
#define BYTE_DEFINED
#include "hyt1331.h"
#include "lrs2365.h"

/*------------------------------------------------------------------*/

int lrs2365_set(int camac_slot, WORD coeff[18])
/**********************************************************************\
   
   Description:

   Load the logic matrix of the LRS2365 with 18 coefficients according
   to the specifications in the manual. After loading, the matrix
   is verified.

\**********************************************************************/
{
WORD i;
WORD data;

  /* Initialize 2365 */
  CAMO(camac_slot, 0, 9, 0);

  /* Set mode control to zero (front-panel input) */
  CAMO(camac_slot, 3, 16, 0);

  /* Write coefficients */
  for (i=0 ; i<18 ; i++)
    CAMO(camac_slot, 0, 16, coeff[i]);

  /* validate coefficients */
  for (i=0 ; i<18 ; i++)
    {
    CAMI(camac_slot, 0, 0, &data);
    if (data != coeff[i])
      {
      cm_msg(MERROR, "lrs2365_set", "Error verifying coefficients: should be %d, read %d.\n", coeff[i], data);
      return 0;
      }
    }

  return SUCCESS;
}
