/********************************************************************\

  Name:         null.c
  Created by:   Stefan Ritt

  Contents:     NULL Device Driver


  Revision history
  -------------------------------------------------------------------
  date         by    modification
  ---------    ---   ------------------------------------------------
  26-MAR-97    SR    created


\********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include "midas.h"

/*---- globals -----------------------------------------------------*/

typedef struct {
  char      port[32];
} NULL_SETTINGS;

NULL_SETTINGS null_settings;

#define NULL_SETTINGS_STR "\
Port = STRING : [32] Nowhere\n\
"

static float *array[10];
static INT   num_channels[10];

/*---- device driver routines --------------------------------------*/

INT null_init(HNDLE hKey, INT index, INT channels)
{
int   status, size;
HNDLE hDB;

  cm_get_experiment_database(&hDB, NULL);

  /* create NULL settings record */
  status = db_create_record(hDB, hKey, "", NULL_SETTINGS_STR);
  if (status != DB_SUCCESS)
    return FE_ERR_ODB;

  size = sizeof(null_settings);
  db_get_record(hDB, hKey, &null_settings, &size, 0);

  /* initialize driver */
  if (array[index] == NULL)
    {
    num_channels[index] = channels;
    array[index] = malloc(sizeof(float) * num_channels[index]);
    }
  else
    {
    num_channels[index] = max(num_channels[index], channels);
    array[index] = realloc(array[index], sizeof(float) * num_channels[index]);
    }
  memset(array[index], 0, sizeof(float) * num_channels[index]);

  return FE_SUCCESS;
}

/*----------------------------------------------------------------------------*/

INT null_exit(INT index)
{
  if (array[index])
    {
    free(array[index]);
    array[index] = NULL;
    }

  return FE_SUCCESS;
}

/*----------------------------------------------------------------------------*/

INT null_set(INT index, INT channel, float value)
{
  if (channel < num_channels[index])
    array[index][channel] = value;

  return FE_SUCCESS;
}

/*----------------------------------------------------------------------------*/

INT null_set_all(INT index, INT channels, float value)
{
INT i;

  for (i=0 ; i<min(num_channels[index], channels) ; i++)
    array[index][i] = value;

  return FE_SUCCESS;
}

/*----------------------------------------------------------------------------*/

INT null_get(INT index, INT channel, float *pvalue)
{
  if (channel < num_channels[index])
    *pvalue = array[index][channel];
  else
    *pvalue = 0.f;

  return FE_SUCCESS;
}

/*----------------------------------------------------------------------------*/

INT null_get_all(INT index, INT channels, float *pvalue)
{
int   i;

  for (i=0 ; i<min(num_channels[index], channels) ; i++)
    pvalue[i] = array[index][i];

  return FE_SUCCESS;
}

/*---- device driver entry point -----------------------------------*/

INT null(INT cmd, ...)
{
va_list argptr;
HNDLE   hKey;
INT     index, channel, status;
float   value, *pvalue;

  va_start(argptr, cmd);
  status = FE_SUCCESS;

  switch (cmd)
    {
    case CMD_INIT:
      hKey = va_arg(argptr, HNDLE);
      index = va_arg(argptr, INT);
      channel = va_arg(argptr, INT);
      status = null_init(hKey, index, channel);
      break;

    case CMD_EXIT:
      index = va_arg(argptr, INT);
      status = null_exit(index);
      break;

    case CMD_SET:
      index = va_arg(argptr, INT);
      channel = va_arg(argptr, INT);
      value   = (float) va_arg(argptr, double);
      status = null_set(index, channel, value);
      break;

    case CMD_SET_ALL:
      index = va_arg(argptr, INT);
      channel = va_arg(argptr, INT);
      value   = (float) va_arg(argptr, double);
      status = null_set_all(index, channel, value);
      break;

    case CMD_GET:
      index = va_arg(argptr, INT);
      channel = va_arg(argptr, INT);
      pvalue  = va_arg(argptr, float*);
      status = null_get(index, channel, pvalue);
      break;
    
    case CMD_GET_ALL:
      index = va_arg(argptr, INT);
      channel = va_arg(argptr, INT);
      pvalue  = va_arg(argptr, float*);
      status = null_get_all(index, channel, pvalue);
      break;
    
    default:
      break;
    }

  va_end(argptr);

  return status;
}

/*------------------------------------------------------------------*/
