/*********************************************************************

  Name:         ps7106.c
  Created by:   Stefan Ritt

  Cotents:      Routines for accessing Phillips Scientific 
                7106 discriminators
                

  Revision history
  ------------------------------------------------------------------
  date        by    modification
  ---------   ---   ------------------------------------------------
  13-NOV-96   SR    created

*********************************************************************/

#include <stdio.h>
#include <dos.h>
#include <math.h>

#include "midas.h"

#define CAMAC_BASE_DEFINED
#define BYTE_DEFINED
#include "hyt1331.h"
#include "ps7106.h"

/*------------------------------------------------------------------*/

int ps7106_set(int camac_slot, double thresh_value)
/********************************************************************\
  Sets the threshold of a Phillips 7106 discriminator to the given
  <thresh_value> in millivolts.
\********************************************************************/
{
unsigned int    threshold;
double          thresh_min = 10;   /* mV */
double          thresh_max = 1024; /* mV */
BOOL            x, q;
WORD            data, value;
char            str[80];

  // Convert threshold value to register value
  threshold = ( thresh_value - 10);
  if ( threshold > 1033 )
    threshold = 1033;

  /* Set remote mode */
  CAMO(camac_slot, 0, SET_REMOTE_7106, 0);

  /* Verify Remote mode */
  data = 0;
  CAMIQ(camac_slot, 0, TEST_REMOTE_7106, &data, &q, &x );
  if (!q)
    {
    cm_msg(MERROR, "ps7106_set", "Error setting remote mode on slot %d.\n", camac_slot);
    return -1;
    }

  /* Write threshold */
  CAMO(camac_slot, 0, WRITE_THRESH_7106, threshold);
  delay(1);

  /* Verify threshold */
  CAMI(camac_slot, 0, READ_THRESH_7106, &data);
  data = data & 0x3FF;
  if ( data != threshold )
    {
    cm_msg(MERROR, "ps7106_set", "Error setting threshold:\
 requested %d (%4.0f mV), read %d.\n", threshold, thresh_value, data);
    return -1;
    }

  /* Measure threshold */
  CAMO(camac_slot, 1, START_ADC_7106, threshold);
  delay(1);
  CAMI(camac_slot, 1, READ_THRESH_7106, &data);
  data = data & 0x3FF;

  if (fabs((double) data - thresh_value) > 2.0)
    {
    cm_msg(MERROR, "ps7106_set", "Error setting threshold on slot %d:\
 requested %1.0lfmV, measured %dmV", camac_slot, thresh_value, data);
    }

  return 0;
}
