/********************************************************************\

  Name:         mspeaker.c
  Created by:   Pierre-Andre Amaudruz

  Contents:     Speaks midas messages


  Revision history
  ------------------------------------------------------------------
  date        by    modification
  --------    ---   ------------------------------------------------
  30-JAN-97   PAA   created

\********************************************************************/

#include "midas.h"
#include "msystem.h"
#include "\provoice\include\fbvspch.h"

LPSTR           lpPhonetics;
LPSPEECHBLOCK   lpSCB;

char *type_name[] = {
  "INFO",
  "ERROR",
  "EXCLUSIVE",
  "USER",
  "DEBUG"
};

/*------------------------------------------------------------------*/

/*----- receive_message --------------------------------------------*/

void receive_message(HNDLE hBuf, HNDLE id, EVENT_HEADER *message)
{
  /* print message */
  printf("%s\n", (char *)(message+1));

  /* skip none talking message */
  if (message->event_id == MT_TALK ||
      message->event_id == MT_USER)
    {
    while (SpeechStatus(lpSCB) != 0)
      ss_sleep(500);

    lpPhonetics = TextToPhonetics(lpSCB, strchr((char *)(message+1),']')+2, 0);
    SpeakPhonetics (lpSCB, lpPhonetics);

    ss_sleep(1000);

//  save phonetics to .wav file
//  WritePhonetics(lpSCB, lpPhonetics, "hello.wav");
    return;
    }

}

/*------------------------------------------------------------------*/

main(int argc, char *argv[])
{
INT    status, i, ch;
char   host_name[NAME_LENGTH];
char   exp_name[NAME_LENGTH];
char   expts[MAX_EXPERIMENT][NAME_LENGTH];

  host_name[0] = 0;
  exp_name[0] = 0;

  /* parse command line parameters */
  for (i=1 ; i<argc ; i++)
    {
    if (argv[i][0] == '-')
      {
      if (i+1 >= argc || argv[i+1][0] == '-')
        goto usage;
      if (argv[i][1] == 'e')
        strcpy(exp_name, argv[++i]);
      else if (argv[i][1] == 'h')
        strcpy(host_name, argv[++i]);
      else
        {
usage:
        printf("usage: mspeaker [-h Hostname] [-e Experiment]\n\n");
        return 0;
        }
      }
    }

  if (equal_ustring(host_name, "local"))
    host_name[0] = 0;

  if (exp_name[0] == 0)
    {
    /* retrieve list of experiments and make selection */
    status = cm_list_experiments(host_name, expts);

    if (status != CM_SUCCESS)
      {
      printf("Cannot connect to host '%s'.\n", host_name);
      return 1;
      }

    if (expts[1][0])
      {
      if (host_name[0])
        printf("Available experiments on server %s:\n", host_name);
      else
        printf("Available experiments on local computer:\n");

      for (i=0 ; expts[i][0] ; i++)
        printf("%ld : %s\n", i, expts[i]);
      printf("Select number: ");
      gets(exp_name);
      i = atoi(exp_name);
      strcpy(exp_name, expts[i]);
      }
    else
      strcpy(exp_name, expts[0]);
    }
  
  /* now connect to server */
  status = cm_connect_experiment(host_name, exp_name, "Speaker", NULL);
  if (status == CM_WRONG_PASSWORD)
    {
    printf("Wrong password\n");
    return 1;
    }
  if (status != CM_SUCCESS)
    {
    printf("Cannot connect to experiment Default on %s\n", host_name);
    return 1;
    }

  cm_msg_register(receive_message);

  printf("Midas Message Talker connected to %s. Press \"!\" to exit\n", 
          host_name[0] ? host_name : "local host");

  //  register to SB
  lpSCB = OpenSpeech(0, 0, "Esnb1k8"); 
	if (lpSCB == NULL)
	  {
		printf("Cannot allocate Speech Control Block\n");
		goto out;
	  }
	else
	  {
	  status = lpSCB->speechStatus;
   	if ((status <= E_SPEECH_ERROR) && (status > W_SPEECH_WARNING))
	    {
		  printf("Open Speech error:%d\n", status);
		  goto out;
	    }
	  }
   
  lpPhonetics = NULL;

  do
    {
    status = cm_yield(1000);
	  if (ss_kbhit())
      {
	    ch = getch();
	    if (ch == 'r')
	      ss_clear_screen();
	    if (ch == '!')
	      break;
	    }
    } while (status != RPC_SHUTDOWN && status != SS_ABORT);
  
out:
  
  CloseSpeech(lpSCB);
  if (lpPhonetics)
	  FreePhoneticsBuffer(lpPhonetics);

  cm_disconnect_experiment();
  return 1;
}
