/*-----------------------------------------------------------------------------
 *  Copyright (c) 1993      TRIUMF Data Acquistion Group
 *  Please leave this header in any reproduction of that distribution
 * 
 *  TRIUMF Data Acquisition Group, 4004 Wesbrook Mall, Vancouver, B.C. V6T 2A3
 *  Email: online@triumf.ca           Tel: (604) 222-1047  Fax: (604) 222-1074
 *         amaudruz@triumf.ca
 * -----------------------------------------------------------------------------
 *  
 *  Description	: replog.c : calls the dlog function for event retrival
 *                see help by invoking replog
 *
 *  Requires 	: dlog.h, dlog.c
 *     
 *  Author:  Pierre-Andre Amaudruz Data Acquisition Group
 * 
 *  Revision History:
 *    $Log: Makefile,v $
 *
 * Revision 1.0  1996/sep/20 Pierre	 Initial revision
 *
 *---------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "midas.h"
#include "msystem.h"
#include "ybos.h"


#define  YBOS_HEADER    1
#define  YBOS_RECORD    2
#define  YBOS_LENGTH    3
#define  YBOS_RAWDEC   11
#define  YBOS_RAWHEX   12
#define  YBOS_YBOSDEC  21
#define  YBOS_YBOSHEX  22
#define  YBOS_YBOSASC  23

DWORD * prevent;
/*------------------------------------------------------------------------------*/
void main (INT argc,char **argv)
{
  INT     bl, action;
  
  if (argc == 2)
    {
      /*- FORMAT INFO -*/
      if (strncmp(argv[1],"fmt",3) == 0)
	{
   	  printf (" -----------------------------------------------------------------------\n");
	  printf (" YBOS Format  : 32 bits word aligned 4 header words\n");
	  printf (" block_size   : (1) LPR Length of Physical Record (0x1ffd exclusive)\n");
	  printf (" header_length: (2) LPH Length of Physical Header (4 exclusive)\n");
	  printf (" block_num    : (3) NPR Number of the Physical Record (start 0)\n");
	  printf (" offset       : (4) Offset count to 1st Logical Record (start 0)\n");
	  printf ("                (5) Next is the Length of the Logical Record (exclusive)\n");
	  printf (" -----------------------------------------------------------------------\n");
	}
    }
  if (argc > 2)
    {
      if (strncmp(argv[2],"h",1)==0)
	action= YBOS_HEADER;
      if (strncmp(argv[2],"r",1)==0)
	action= YBOS_RECORD;
      if (strncmp(argv[2],"l",1)==0)
	action= YBOS_LENGTH;
      if (strncmp(argv[2],"ed",2)==0)
	action= YBOS_RAWDEC;
      if (strncmp(argv[2],"ex",2)==0)
	action= YBOS_RAWHEX;
      if (strncmp(argv[2],"yd",2)==0)
	action= YBOS_YBOSDEC;
      if (strncmp(argv[2],"yx",2)==0)
	action= YBOS_YBOSHEX;
      if (strncmp(argv[2],"ya",2)==0)
	action= YBOS_YBOSASC;
      
      if (argc >= 3)
	{
	  bl = -1;
	  if (argc == 4)
	    bl = atoi(argv[3]);
	  /* open data file */
	  if (ybos_open_datafile (argv[1]) != RDLOG_SUCCESS)
	    exit (-1);
	  /* skip record if necessary */
	  if (bl > -1)
	    if (ybos_skip_physrec(bl) != RDLOG_SUCCESS)
	      exit (-1);
	  
	  /* get physical record header */
	  if (bl == -1)
	    {
	      if ((action > 0) && (action < YBOS_LENGTH))
		{
		  while (ybos_physrec_get() == RDLOG_SUCCESS)
		    {
		      switch (action)
			{
			case YBOS_HEADER :
			  ybos_physhead_display();
			  break;
			case YBOS_RECORD :
			  ybos_physrec_display();
			  break;
			}
		    }
		}
	      else if ((action > YBOS_RECORD) && (action <= YBOS_YBOSASC))
		{
		  while (ybos_event_get(bl, prevent) == RDLOG_SUCCESS)
		    {
		      ybos_swap_event(prevent);
		      switch (action)
			{
			case YBOS_LENGTH :
			  ybos_evtlen_display();
			  break;
			case YBOS_RAWDEC :
			case YBOS_RAWHEX :
			case YBOS_YBOSDEC :
			case YBOS_YBOSHEX :
			case YBOS_YBOSASC :
			  display_any_event(prevent, FORMAT_YBOS, action/10, action%10);
			  break;
			}
		    }
		}
	    }
	  else
	    {
	      if ((action > 0) && (action < YBOS_LENGTH))
		{
		  switch (action)
		    {
		    case YBOS_HEADER :
		      ybos_physhead_display();
		      break;
		    case YBOS_RECORD :
		      ybos_physrec_display();
		      break;
		    }
		}
	      else if ((action > YBOS_RECORD) && (action <= YBOS_YBOSASC))
		{
		  while (ybos_event_get(bl, prevent) == RDLOG_SUCCESS)
		    {
		      ybos_swap_event(prevent);
		      switch (action)
			{
			case YBOS_LENGTH :
			  ybos_evtlen_display();
			  break;
			case YBOS_RAWDEC :
			case YBOS_RAWHEX :
			case YBOS_YBOSDEC :
			case YBOS_YBOSHEX :
			case YBOS_YBOSASC :
			  display_any_event(prevent, FORMAT_YBOS, action/10, action%10);
			  break;
			}
		    }
		}
	    }
	  /* close data file */   
	  ybos_close_datafile();
	  exit (0);
	}
    }
  printf ("replog <file_name> h   :header          [block#]>\n");
  printf ("       <file_name> r   :record          [block#]>\n");
  printf ("       <file_name> l   :evt_length      [block#]>\n");
  printf ("       <file_name> ex  :event     (hex) [block#]>\n");
  printf ("       <file_name> ed  :event     (dec) [block#]>\n");
  printf ("       <file_name> yx  :ybos event(hex) [block#]>\n");
  printf ("       <file_name> yd  :ybos event(dec) [block#]>\n");
  printf ("       <file_name> ya  :ybos event(asc) [block#]>\n");
  printf ("       <fmt>          :format info\n");
  exit(0);
}
